
###############################################################################
# A PyIPC example. This code IS NOT PART of PyIPC, and does not fall under    #
# it's licence. Therefore, you can do whatever you want with this code.       #
#                                                                             #
# This code is in the public domain.                                          #
###############################################################################

from pyipc import *

#----------------------------------------------------------------------------
class SharedObject: 
    """
    Here is an example of how you can use a Shared Memory Block and a Semaphore
    to share a python object between two python processes.
    
    This is just an example, and, a slow one. To make this fast, you would
    use C code to keep the object in Shared Memory.  For my purposes, this has 
    worked though, and you might find it usefull.
    
    This module makes some assumptions that could be made more robust if someone
    were so inclined. It assumes that pickled objects will not be greter than 
    4094 bytes. It also assumes that people using this class are decent 
    programmers, and will call wait() and signal() correctly. So, unless you are
    writing all processes using this code, there could be a security risk. To 
    minimize security issues, the permissions on both the semaphore and the 
    share memory is 0600. This code also does not allow writing a 'None'
    object to memory on the first write. You could easily change the interface
    if you like. Ain't open source great?

    Note that there is no protection against a process which does not hold
    write acces to the object calling the signal method (thereby unlocking 
    the object).  Use good coding practices, and you'll be OK. If you want
    to make it more robust, send me the changes, and I will include them and
    give you credit.
    """
    #------------------------------------------------------------------------
    def __init__(self, semKey, shmKey, obj=None):
        """
        __init__(semKey=1, shmKey=1, obj=None)

        This constructor has two modes.  Either the obj is None, or it has
        a real instance in it.  
        
        If it is None, it is assumed that you are binding to a 
        SharedObject that already exists.  In that case, the semKey and shmKey
        values are strict, meaning that if those ipc components can not be
        obtained, an error will be raised.

        If the object is not none, then It is assumed that you are creating 
        a new SharedObject, and the semKey and shmKey values are used to 
        create exclusive ipc components.  If an exclusive SharedMemory object
        or Semaphore object can not be obtained, then the value + 1 is tried.
        In other words, this constructor does its best to get a Shared Memory
        block and Semaphore group for you.  Once they are obtained, you can 
        get their keys with the keys() method. Note that only 128 alternate 
        value will be attempted.  If there is not success in those tries, then
        a MemoryError will be raised.
        """

        self._mem = None
        self._sem = None
        
        if obj!=None:
            i = 0
            while i < 128:
                try:
                    self._mem = SharedMemory( shmKey , SHMLBA, 
                                              0600 | IPC_EXCL | IPC_CREAT)
                    i = 128
                except:
                    shmKey += 1
                    i += 1

            i = 0
            while i < 128:
                try:
                    self._sem = SemaphoreGroup( semKey, 1,
                                                0600 | IPC_EXCL | IPC_CREAT)
                    i = 128
                except:
                    semKey += 1
                    i += 1

            if self._mem and self._sem:
                # Success in creating exclusive ipc components. write obj.
                str = obj2string(obj)
                self._mem.writeUInt(0,len(str))
                self._mem[SIZE_LONG] = str

        else:
            try:
                self._mem = SharedMemory  ( shmKey , SHMLBA , 0600 )
                self._sem = SemaphoreGroup( semKey , 1 , 0600 , 0  )
            except:
                pass

        if not self._mem or not self._sem:
            raise "SharedObject: incorrect id given in constructor."

        self._memKey = shmKey
        self._semKey = semKey

    #------------------------------------------------------------------------
    def signal(self, obj):
        """
        signal(object)

        Write the object to the sharedMemory, and release the key.
        """
        str = obj2string(obj)
        self._mem.writeUInt(0,len(str))
        self._mem[SIZE_LONG] = str
        self._sem.signal()
    
    #------------------------------------------------------------------------
    def wait(self):
        """
        wait()

        Obtain write permission on the object.  The object which is being
        saved in memory will be returned. Note that the object returned will
        not have the same reference value as the object passed to the 
        constructor since this method actually returns a copy of that object.
        """
        self._sem.wait()
        length = self._mem.readUInt(0)
        str = self._mem[SIZE_LONG: SIZE_LONG + length]
        return string2obj(str)

    #------------------------------------------------------------------------
    def keys(self):
        """
        keys()

        Get a tuple of length two with the Semaphore key and the Shared 
        Memory key respectively.  You will need these if you create the 
        shared object in this process, and you want to pass the values to 
        another process to open communication there.
        """
        return (self._semKey, self._shmKey)

#----------------------------- END SharedObject.py --------------------------
