#Class Dispatcher
#Author Wong Yan
#Class SingleThread, Multithread, LockedIterator
#Author: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/203871
#Date 06/18/2004

import threading
import time
import types

# Pool of threads: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/203871
class SingleThread( threading.Thread ):
    """Definition of a single Thread
    """
    def  __init__( self, target, args=() ):
        if type( args ) <> types.TupleType:
            args = (args,)
        threading.Thread.__init__( self, target=target, args=args )

class LockedIterator:
    """ Avoid concurrency on the iterator
    """
    def __init__( self, iterator ):
        self._lock     = threading.Lock()
        self._iterator = iterator

    def __iter__( self ):
        return self

    def next( self ):
        try:
            self._lock.acquire()
            return self._iterator.next()
        finally:
            self._lock.release()

class MultiThread:
    """Several threads
    """
    def __init__( self, function, argsVector, maxThreads=5 ):
        self._function     = function
        self._argsIterator = LockedIterator( iter( argsVector ) )
        self._threadPool   = []
        for i in range( maxThreads):
            self._threadPool.append( SingleThread( self._tailRecurse ) )

    def _tailRecurse( self ):
        for args in self._argsIterator:
            self._function( args ) 

    def start( self ):
        for thread in self._threadPool:
            time.sleep( 0 ) # necessary to give other threads a chance to run
            thread.start()

    def join( self, timeout=None ):
        for thread in self._threadPool:
            thread.join( timeout )

class EDispatchError(Exception):
    """Mistake during the process of a query
    """
    def __str__(self):
        return "An error occurs during the Dispatcher process: "

	
class AbstractDispatcher:
    """ Subclass this one to create your own dispatcher class
        a Dispatcher class must have a execute method
    """
    def __init__(self, mobyContentXML, Invocator, invocatorParameters, CommandBuilder, formatter=None, workdir="/tmp"):
        """ dispatch a MobyContent into several queries
            need the class of Invocator, optional parameters of Invocation (in a InvocatorParameters class object)
            a Command builder, a function that changes the queries into Invocator Parameters
            formatter: a function that transforms raw results from the Invocator into Moby Objects
            workdir: the session directory
        """
	import base64, binascii, re

	from bioMoby import MobyUnmarshaller
        um=MobyUnmarshaller()
        
	self._isbase64=False
	self._workdir=workdir
	
        try:
	    #Do some cleaning on the XML Document before proceeding...
	    #Remove all \t|\n from the <a>\n\t</a>
	    r=re.compile("\>[\n|\t]+\<")
            mc=um.loads("><".join(r.split(mobyContentXML)))
        except:
            try:
		mc=um.loads(base64.decodestring(mobyContentXML))
	        self._isbase64=True
	    except binascii.Error:
	        raise  EDispatchError, "Invalid types in the moby Content object"

        self.queryData=mc.queryData
        self.invocator=Invocator
	self._invocatorParameters=invocatorParameters
        self.commandBuilder=CommandBuilder
        self.formatter=formatter
	
	del MobyUnmarshaller, re

    def execute(self):
	""" Override this method
	"""
	pass

	
class SimpleDispatcher(AbstractDispatcher):
    """ A simple dispatcher, without execute method
    """
    def __init__(self, mobyContentXML, Invocator, InvocatorParameters, CommandBuilder, formatter=None, workdir="/tmp"):
	AbstractDispatcher.__init__(self, mobyContentXML, Invocator, InvocatorParameters, CommandBuilder, formatter, workdir)
	
	from bioMoby import MobyObject
	import random, os, sys
	
	if 'JOBSESSION' in self.queryData.keys():
	    self._sessionid=self.queryData['JOBSESSION'][0].id
	else:
	    self._sessionid=`random.randint(0,sys.maxint-1)`
	    
	    while "session"+self._sessionid in os.listdir(self._workdir):
		self._sessionid=`random.randint(0,sys.maxint-1)`
		
            os.mkdir(self._workdir+"/session"+self._sessionid)
	
	os.chdir(self._workdir+"/session"+self._sessionid)
        self.answers={'JOBSESSION':[MobyObject(namespace="SESSIONID", id=self._sessionid)]}

			
class PBSDispatcher(SimpleDispatcher):
    """Dispatcher for PBS jobs
    """

    def __init__(self, mobyContentXML, InvocatorParameters, CommandBuilder, formatter=None, workdir="/tmp"):
	""" PBS Dispatcher work with PBSInvocator
	    it uses a session directory, a session file (where the queries IDs are stored)
	    when the execute method is call with an session id, it reads all entries from the session file and
	    make a qstat on each ids.
	"""
	from bioMoby.webservice import PBSInvocator
	
	SimpleDispatcher.__init__(self, mobyContentXML, PBSInvocator, InvocatorParameters, CommandBuilder, formatter, workdir)
    
    def _getResults(self):
	""" Do a qstat and check availability of results
	"""
	import os, shutil
	from bioMoby import GeneralInvocator
	
	answers={}
	outputfiles={}
	queryKeys=[]
	
	#Read the sessions file, in which there are all queries identifiers
	
	fp=file("sessions","r")
	for queryInformation in fp.readlines():
	    #for each queries, retrieve the qsub ids, and associated output files.
	    #delete the CR at the end of the line before proceding
	    queryInformation=queryInformation[:-1] 
	    #Get the IDs
	    qsubId=queryInformation.split(";")[1].split(".")[0]
	    qk=queryInformation.split(";")[0]
	    #Get the query key
	    queryKeys.append(qk)
            #get the associated Outputfiles
	    outputfiles[qk]=queryInformation.split(";")[2].split("::")

	    #Do a qstat and check if the job is still running or on queue
	    i, o, e = os.popen3("qstat %s"%qsubId)

	    orl=o.readlines()
	    erl=e.readlines()
	    
	    if len(orl)>0:
		#if one job is not finished, exit the _getResults function
		#The dispatcher will give all the results or none 
		return None
	
	fp.close()
	
	# At this step, all jobs are done
	tdir=self._workdir+"/session"+self._sessionid
	
	
	for querykey in queryKeys:
	    #check the results-query.out:
	    if "results-"+querykey+".out" in os.listdir(tdir):
                fp=file(tdir+"/results-"+querykey+".out")
	        s=fp.readlines()
	        fp.close()

                if len(s)==0:
	            os.remove(tdir+"/results-"+querykey+".out")
	        else:
		    if len(outputfiles.keys())>0:
		        shutil.copy(tdir+"/results-"+querykey+".out", tdir+"/results-"+querykey+"/results.out")

            for f in outputfiles[querykey]:
	        ruserdir=tdir+"/results-"+querykey
                # if it is a file then retrieve the file
                if os.path.isfile(f):
                    aTemp=f.split('/')
                    shutil.copy(f, ruserdir+"/"+aTemp[len(aTemp)-1])
                    os.remove(f)
                elif os.path.isdir(f):
                    #if it is a directory retrieves all results in a collection
                    for fs in os.listdir(f):
                        shutil.copy(f+"/"+fs, ruserdir+"/"+fs)
                        os.remove(f+"/"+fs)
                    os.removedirs(f)

	#retrieve all the results
        try:
	    l=os.listdir(tdir)
	except:
	    l=[]
	
	if len(l)>0:
            for querykey in l:
		if querykey[:8]=="results-":
		    aQueryName=querykey[8:]
	            gi=GeneralInvocator(aQueryName, userdir=tdir)
	            answers[aQueryName]=[gi.getResults()]

                    if self.formatter != None:
                        answers[aQueryName]=self.formatter(answers[aQueryName][0])
			
            if len(answers.keys())>0:
	        self.answers=answers
		
                try:
	            #Clean the session directory
	            shutil.rmtree(tdir)
		    os.removedirs(tdir)
	        except:
		    pass
	
	del GeneralInvocator,os, shutil

	
    def execute(self):
	"""Execute the command in a PBS script using the PBSInvocator
	"""
	from bioMoby import MobyContent
	import base64

        if 'JOBSESSION' in self.queryData.keys():
	    self.answers=self.queryData

	    self._getResults()
	else:
	    #we need to store informations in a file
	    #query id, qsubid, outputfiles
	    fp=file("sessions","w")
	    for queryKey in self.queryData.keys():
		query=self.queryData[queryKey]
	        r=self.invocator(queryKey, self.commandBuilder(query), fp, userdir=self._workdir+"/session"+self._sessionid)
		r.setParameters(self._invocatorParameters)
		r.execute()
	    fp.close()

    	if self._isbase64:
	    return base64.encodestring(str(MobyContent(self.answers)))
	    
        return str(MobyContent(self.answers))

class Dispatcher(AbstractDispatcher):
    """All Queries in the MobyContent Object are splitted and executed
    """
    def __init__(self, mobyContentXML, Invocator, CommandBuilder, formatter=None, workdir="/tmp"):
        """The Object needs a MobyContent XML string
            an Invocator Class and a commandBuilder function
            The commandBuilder function shall return  a command, arguments, and a set of temporary files
            [queryData]-->("commandName","commandArgs", ["tempfiles"])
        """
	AbstractDispatcher.__init__(self, mobyContentXML, Invocator, None, CommandBuilder, formatter, workdir="/tmp")
	
	import os, random, sys, pyipc
	from bioMoby import MobyObject
	
	#Create a Jobsession ID, a semaphore and a session directory
	if 'JOBSESSION' in self.queryData.keys():
	    self._sessionid=self.queryData['JOBSESSION'][0].id
	    self._sem=pyipc.SemaphoreGroup(int(self._sessionid))
	else:
	    self._sessionid=`random.randint(0,sys.maxint-1)`
	    
	    while "session"+self._sessionid in os.listdir(self._workdir):
		self._sessionid=`random.randint(0,sys.maxint-1)`
		
            os.mkdir(self._workdir+"/session"+self._sessionid)
            self._sem=pyipc.SemaphoreGroup(int(self._sessionid))
            self._sem.wait()
	
	#By default, we return the Session ID
        self.answers={'JOBSESSION':[MobyObject(namespace="SESSIONID", id=self._sessionid)]}

	#Create a lock for synchronisation for the Multithreaded execution
	self._lock=threading.RLock()
        
	del os, random, sys, pyipc

    def _execute(self, queryKey):
        """ Execute a single query, store the result in the answers attribute
        """
        from bioMoby import MobyObject, MobyString

        query=self.queryData[queryKey]
	
	result=[MobyString("The "+queryKey+" has not been processed")]
	
	#Execute the query
        try:
            r=self.invocator(queryKey, self.commandBuilder(query), userdir=self._workdir+"/session"+self._sessionid)
            result=[r.execute()]
            time.sleep(0)
        except:
	    #If there is a problem, we return the generated exception
	    import traceback
	    fp=file(self._workdir+"/session"+self._sessionid+"/results-"+queryKey,"w")
	    fp.write("The query: "+queryKey+" has not been processed")
	    traceback.print_exc(file=fp)
	    fp.close()

	#Avoir concurrency writing on self.answers
        self._lock.acquire()
        self.answers[queryKey]=result
        self._lock.release()

	    
        del MobyObject, MobyString

    def _fetchAnswers(self):
	""" Fetch answers if they exists
	"""
        from bioMoby import GeneralInvocator
        import os, shutil, pyipc

	answers={}

	ruserdir="session"+self._sessionid
        try:
	    l=os.listdir(self._workdir+"/"+ruserdir)
	except:
	    l=[]
	
	if len(l)>0:
	    #There is a session directory
            for querykey in l:
		if querykey[:8]=="results-":
		    aQueryName=querykey[8:]
	            gi=GeneralInvocator(aQueryName ,userdir=self._workdir+"/"+ruserdir)
	            answers[aQueryName]=[gi.getResults()]

                    if self.formatter != None:
                        answers[aQueryName]=self.formatter(answers[aQueryName][0])
			
            if len(answers.keys())>0:
	        self.answers=answers
		
                try:
	            #Clean the session directory
	            shutil.rmtree(self._workdir+"/"+ruserdir)
		    os.removedirs(self._workdir+"/"+ruserdir)
	        except:
		    pass

	#Remove the IPC
	pyipc.removeIPC(self._sem)
	
	del GeneralInvocator,os, shutil

	
    def _getSem(self):
	try:
	    #Wait for answers of all queries
	    self._sem.wait()
	    self._fetchAnswers()
        except:
	    pass
	
    def execute(self, timeout=120):
        """Execute the queries with a pool of threads, store the results in a mobyContent Object
        """
	import os, pyipc, signal
	from bioMoby import MobyContent
	
	#Do a fetchAnswers if the user has a sessionid
	if 'JOBSESSION' in self.queryData.keys():
	    self.answers=self.queryData

	    t=threading.Thread(target=self._getSem)
	    t.start()
	    t.join(timeout)

	    if self._isbase64:
	        import base64
	        return base64.encodestring(str(MobyContent(self.answers)))
	    
            return str(MobyContent(self.answers))
        else:
	    #A MobyContent without jobsession id has been received
	    try:
		#make a fork
	        pid=os.fork()
	    except:
		raise Exception, "Cannot treat any further request, please retry"

	    if pid>0:
		#try to get the results
#		os.wait()
	        t=threading.Thread(target=self._getSem)
	        t.start()
	        t.join(timeout)

                if self._isbase64:
	            import base64
	            return base64.encodestring(str(MobyContent(self.answers)))
	    
                return str(MobyContent(self.answers))		
	    else:
                #do second fork
                try: 
                    pid = os.fork()
		except:
		    raise Exception, "Cannot treat any further request, please retry"
		    
                if pid>0:
		    #it it is too long, kill the process
		    time.sleep(1.05*timeout)
                    os.kill(os.getppid(), signal.SIGKILL)
		    os._exit(0)
		else:
		    #execute the treatment
                    os.setpgrp()
		    os.chdir(self._workdir+"/session"+self._sessionid)
	            mt=MultiThread(self._execute, self.queryData.keys())
                    mt.start()
                    mt.join()

		    try:
	                self._sem.signal()
                    finally:
			if os.getppid()>1:
                  	    try:
			        os.kill(os.getppid(), signal.SIGKILL)
		            except:
			        pass
			    
		        os._exit(0)